package org.bitbucket.bradleysmithllc.etlunit.feature.sql_server_database;

import javax.inject.Inject;
import javax.inject.Named;

import net.sourceforge.jtds.jdbc.Driver;
import org.bitbucket.bradleysmithllc.etlunit.Log;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.TestExecutionError;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.*;
import org.bitbucket.bradleysmithllc.etlunit.feature.database.db.Table;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

public class SqlServerDatabaseImplementation extends BaseDatabaseImplemenation
{
	private Log applicationLog;
	private RuntimeSupport runtimeSupport;
	private DatabaseFeatureModule databaseFeatureModule;

	@Inject
	public void setRuntimeSupport(RuntimeSupport runtimeSupport)
	{
		this.runtimeSupport = runtimeSupport;
	}

	@Inject
	public void setDatabaseFeatureModule(DatabaseFeatureModule databaseFeatureModule)
	{
		this.databaseFeatureModule = databaseFeatureModule;
	}

	private Map<String, Table> identityStatus = new HashMap<String, Table>();

	@Override
	public void prepareConnectionForInsert(Connection connection, Table target, DatabaseConnection dc, String mode) throws Exception
	{
		if (mode == null)
		{
			mode = "[[DEFAULT]]";
		}

		String identityKey = dc.getId() + "." + mode;

		if (target.hasIdentityColumns())
		{
			applicationLog.info("Enabling identity insert for " + target.getQualifiedName());

			Table currIdent = identityStatus.get(identityKey);
			Statement st = connection.createStatement();

			if (currIdent != null)
			{
				String sql = "SET IDENTITY_INSERT " + escapeQualifiedIdentifier(currIdent) + " OFF";

				try
				{
					st.execute(sql);
				}
				catch (SQLException exc)
				{
					applicationLog.severe("Error disabling identity insert for " + currIdent.getQualifiedName() + " using sql: " + sql, exc);
				}
			}

			String sql = "SET IDENTITY_INSERT " + escapeQualifiedIdentifier(target) + " ON";

			try
			{
				st.execute(sql);
				identityStatus.put(identityKey, target);
			}
			catch (SQLException exc)
			{
				applicationLog.severe("Error enabling identity insert for " + target.getQualifiedName() + " using sql: " + sql, exc);
			}
			finally
			{
				st.close();
			}
		}
		else
		{
			applicationLog.info("Identity insert not required for " + target.getQualifiedName());
		}
	}

	@Inject
	public void setLogger(@Named("applicationLog") Log log)
	{
		this.applicationLog = log;
	}

	public String getImplementationId()
	{
		return "sql-server";
	}

	public Object processOperationSub(operation op, OperationRequest request) throws UnsupportedOperationException
	{
		switch (op)
		{
			case createDatabase:
				doCreateDatabase(request.getInitializeRequest());
				break;

			case materializeViews:
				doMaterializeViews(request.getInitializeRequest());
				break;
		}

		return null;
	}

	public String getDefaultSchema(DatabaseConnection dc, String mode)
	{
		return "DBO";
	}

	private void doMaterializeViews(InitializeRequest initializeRequest)
	{
		URL url_materialize_views = getClass().getResource("/sqlserver_materialize_views.vm");

		Map<String, String> map = new HashMap<String, String>();
		DatabaseConnection databaseConnection = initializeRequest.getConnection();
		String mode = initializeRequest.getMode();

		try
		{
			// materialize the views into tables
			String materializeText = IOUtils.readURLToString(url_materialize_views);

			String script = VelocityUtil.writeTemplate(materializeText, map);

			String scriptName = "Materialize_Views_" + getDatabaseName(databaseConnection, mode) + ".sql";

			File scriptFile = runtimeSupport.createGeneratedSourceFile(getImplementationId(), scriptName);

			IOUtils.writeBufferToFile(scriptFile, new StringBuffer(script));

			executeScript(script, databaseConnection, mode, false);
		}
		catch (Exception e)
		{
			// TODO:  Get the appropriate exceptions added to this interface method
			throw new IllegalArgumentException(e);
		}
	}

	private void executeScript(String script, DatabaseConnection databaseConnection, String mode, boolean master) throws IOException, TestExecutionError
	{
		final SQLAggregator sqlagg = databaseFeatureModule.resolveSQLRef(script);

		jdbcClient.useStatement(databaseConnection, mode, new JDBCClient.StatementClient()
		{
			public void connection(Connection conn, Statement st, DatabaseConnection connection, String mode, int id) throws Exception
			{
				SQLAggregator.Aggregator statementAggregator = sqlagg.getStatementAggregator();
				while (statementAggregator.hasNext())
				{
					st.execute(statementAggregator.next().getLine());
				}
			}
		}, master ? SYSADMIN : DATABASE);
	}

	private void doCreateDatabase(InitializeRequest initializeRequest)
	{
		URL url_kill = getClass().getResource("/sqlserver_kill_db.vm");
		URL url_create = getClass().getResource("/sqlserver_create_db.vm");

		try
		{
			String killText = IOUtils.readURLToString(url_kill);

			Map<String, String> map = new HashMap<String, String>();
			DatabaseConnection databaseConnection = initializeRequest.getConnection();
			String mode = initializeRequest.getMode();
			map.put("databaseLogin", getLoginName(databaseConnection, mode));
			map.put("databaseName", getDatabaseName(databaseConnection, mode));
			map.put("databaseUsername", getLoginName(databaseConnection, mode));
			map.put("databasePassword", getPassword(databaseConnection, mode));

			String script = VelocityUtil.writeTemplate(killText, map);

			String scriptName = "Kill_Database_" + getDatabaseName(databaseConnection, mode) + ".sql";

			File scriptFile = runtimeSupport.createGeneratedSourceFile(getImplementationId(), scriptName);

			IOUtils.writeBufferToFile(scriptFile, new StringBuffer(script));

			// execute the script through sqlcmd
			executeScript(script, databaseConnection, mode, true);

			String createText = IOUtils.readURLToString(url_create);

			script = VelocityUtil.writeTemplate(createText, map);

			scriptName = "Create_Database_" + getDatabaseName(databaseConnection, mode) + ".sql";

			scriptFile = runtimeSupport.createGeneratedSourceFile(getImplementationId(), scriptName);

			IOUtils.writeBufferToFile(scriptFile, new StringBuffer(script));

			executeScript(script, databaseConnection, mode, true);
		}
		catch (Exception e)
		{
			// TODO:  Get the appropriate exceptions added to this interface method
			throw new IllegalArgumentException(e);
		}
	}

	private void doDropConstraints(InitializeRequest prepareRequest)
	{
		URL url = getClass().getResource("/sqlserver_drop_constraints.vm");

		try
		{
			String purgeText = IOUtils.readURLToString(url);
			DatabaseConnection databaseConnection = prepareRequest.getConnection();
			String mode = prepareRequest.getMode();
			String databaseName = getDatabaseName(databaseConnection, mode);

			String scriptName = "Drop_Constraints_" + databaseName + ".sql";

			File scriptFile = runtimeSupport.createGeneratedSourceFile(getImplementationId(), scriptName);

			IOUtils.writeBufferToFile(scriptFile, new StringBuffer(purgeText));

			// execute the script through sqlcmd
			executeScript(purgeText, databaseConnection, mode, false);
		}
		catch (Exception e)
		{
			// TODO:  Get the appropriate exceptions added to this interface method
			throw new IllegalStateException("", e);
		}
	}

	@Override
	public String getJdbcUrl(DatabaseConnection dc, String mode, int id)
	{
		String serverPort = dc.getServerPort() != -1 ? (":" + String.valueOf(dc.getServerPort())) : "";

		if (id == SYSADMIN)
		{
			return "jdbc:jtds:sqlserver://" + dc.getServerName() + serverPort + "/master";
		}
		else
		{
			return "jdbc:jtds:sqlserver://" + dc.getServerName() + serverPort + "/" + getDatabaseName(dc, mode);
		}
	}

	@Override
	public Class getJdbcDriverClass()
	{
		return Driver.class;
	}

	@Override
	public boolean isTableTestVisible(DatabaseConnection dc, String mode, Table table) {
		String schemaName = table.getSchema().getName();
		boolean sysSchema = schemaName != null && (schemaName.equals("SYS") || schemaName.equals("INFORMATION_SCHEMA"));

		return
			super.isTableTestVisible(dc, mode, table)
			&& !sysSchema;
	}

	@Override
	public boolean restrictToOwnedSchema(DatabaseConnection dc) {
		return false;
	}
}